import cv2
import json
import numpy as np
import onnxruntime as ort

from ..utils import download_file, get_providers


def get_input_points(prompt):
    prompt = json.loads(prompt)
    points, labels = [], []
    for mark in prompt:
        if mark["type"] == "point":
            points.append(mark["data"])
            labels.append(mark["label"])
        elif mark["type"] == "rectangle":
            points.append([mark["data"][0], mark["data"][1]])
            points.append([mark["data"][2], mark["data"][3]])
            labels.append(2)
            labels.append(3)
    return np.array(points), np.array(labels)


class SAM:

    def __init__(self):
        self.target_size = 1024
        self.input_size = (684, 1024)
        encoder_src = download_file('multiple/models/mobile_sam.encoder.onnx')
        decoder_src = download_file('multiple/models/mobile_sam.decoder.onnx')
        self.encoder = ort.InferenceSession(encoder_src, providers=get_providers())
        self.decoder = ort.InferenceSession(decoder_src, providers=get_providers())

    def encode(self, img):
        # TODO: Refactor to use resize
        scale_x = self.input_size[1] / img.shape[1]
        scale_y = self.input_size[0] / img.shape[0]
        scale = min(scale_x, scale_y)

        transform_matrix = np.array([[scale, 0, 0],
                                     [0, scale, 0],
                                     [0, 0, 1]])

        size = (self.input_size[1], self.input_size[0])
        img = cv2.warpAffine(img, transform_matrix[:2], size, flags=cv2.INTER_LINEAR)

        encoder_input_name = self.encoder.get_inputs()[0].name
        encoder_inputs = {encoder_input_name: img.astype(np.float32)}
        encoder_output = self.encoder.run(None, encoder_inputs)
        self.image_embedding = encoder_output[0]
        return self.image_embedding
    
    def decode(self, image_embedding, input_points, input_labels):
        onnx_coord = np.concatenate([input_points, np.array([[0.0, 0.0]])], axis=0)[None, :, :]
        onnx_label = np.concatenate([input_labels, np.array([-1])], axis=0)[None, :].astype(np.float32)
        onnx_coord = np.concatenate([onnx_coord, np.ones((1, onnx_coord.shape[1], 1), dtype=np.float32)], axis=2)
        onnx_coord = onnx_coord[:, :, :2].astype(np.float32)

        onnx_mask_input = np.zeros((1, 1, 256, 256), dtype=np.float32)
        onnx_has_mask_input = np.zeros(1, dtype=np.float32)

        decoder_inputs = {"image_embeddings": image_embedding,
                          "point_coords": onnx_coord,
                          "point_labels": onnx_label,
                          "mask_input": onnx_mask_input,
                          "has_mask_input": onnx_has_mask_input,
                          "orig_im_size": np.array(self.input_size, dtype=np.float32)}
        masks, _, _ = self.decoder.run(None, decoder_inputs)
        return masks[0]

    def predict(self, img, prompt="[]"):
        height, width = img.shape[:2]

        scale_x = self.input_size[1] / img.shape[1]
        scale_y = self.input_size[0] / img.shape[0]
        scale = min(scale_x, scale_y)

        transform_matrix = np.array([[scale, 0, 0],
                                     [0, scale, 0],
                                     [0, 0, 1]])
        
        if self.image_embedding is None:
            self.image_embedding = self.encode(img)

        input_points, input_labels = get_input_points(prompt)
        input_points = input_points * scale
        masks = self.decode(self.image_embedding, input_points, input_labels)

        inv_transform_matrix = np.linalg.inv(transform_matrix)
        mask = np.zeros((height, width), dtype=np.uint8)
        for m in masks:
            m = cv2.warpAffine(m, inv_transform_matrix[:2], (width, height), flags=cv2.INTER_LINEAR)
            mask[m > 0.0] = 255
        return mask



import os
import cv2
import numpy as np
from PIL import Image

from ..inference.faces import Retinaface


def spectral_residual_mono(gray):
    imfft = cv2.dft(gray.astype(np.float32), flags=cv2.DFT_COMPLEX_OUTPUT)
    magnitude, phase = cv2.cartToPolar(imfft[:, :, 0], imfft[:, :, 1])
    log_amplitude = cv2.log(magnitude + 1e-10)
    exp_magnitude = cv2.exp(log_amplitude - cv2.blur(log_amplitude, (3, 3)))
    spectral_residual = np.dstack(
        [exp_magnitude * np.cos(phase), exp_magnitude * np.sin(phase)])
    salmap = cv2.idft(
        spectral_residual, flags=cv2.DFT_SCALE | cv2.DFT_REAL_OUTPUT)**2
    return salmap


def spectral_residual(img):
    """Computes a salience map with the spectral residual method."""
    rimg = cv2.resize(img, (64, 64), interpolation=cv2.INTER_AREA)
    if img.ndim == 3:
        rimg = cv2.cvtColor(rimg, cv2.COLOR_RGB2LAB)
        ss = [spectral_residual_mono(rimg[:, :, c]) for c in range(3)]
        salmap = np.sum(np.dstack(ss), axis=2)
    else:
        salmap = spectral_residual_mono(rimg)
    salmap = cv2.GaussianBlur(salmap, (11, 11), 2.5, dst=salmap)
    salmap = cv2.resize(salmap, (img.shape[1], img.shape[0]))
    return cv2.normalize(salmap, None, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)


def draw_ellipses(img, rects, color=255, thickness=-1):
    for x, y, w, h in rects:
        cv2.ellipse(img, (x+(w-1)//2, y+(h-1)//2), ((w-1)//2, (h-1)//2), 0, 0, 360, color, thickness)
    return img


def combine_maps(map1, map2, alpha=0.3):
    cmap = cv2.addWeighted(map1, 1-alpha, map2, alpha, 0)
    return cv2.normalize(cmap, cmap, 0, 255, cv2.NORM_MINMAX, cv2.CV_8U)


class Faces: 
    def __init__(self):
        self.detector = Retinaface()

    def detect(self, img):
        results = self.detector.detect_faces(img)
        return [result['box'] for result in results]


class Saliency:
    def __init__(self):
        self.center = cv2.imread(os.path.join(os.path.dirname(__file__), 'center.jpg'))[:, :, 0]

    def faces_map(self, salmap, rects, alpha):
        if len(rects):
            facemap = draw_ellipses(np.zeros(salmap.shape, dtype=np.uint8), rects, 255)
            facemap = cv2.GaussianBlur(facemap.astype(np.float32) / 255, (249, 249), 30)
            facemap = cv2.normalize(facemap, None, alpha=0, beta=255, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_8U)
            salmap = combine_maps(salmap, facemap, alpha)
        return salmap

    def center_model(self, salmap, alpha):
        size = salmap.shape[1], salmap.shape[0]
        modmap = cv2.resize(self.center, size)
        salmap = combine_maps(salmap, modmap, alpha)
        return salmap

    def predict(self, img, faces=[]):
        salmap = spectral_residual(img)
        salmap = self.faces_map(salmap, faces, 0.35)
        salmap = self.center_model(salmap, 0.05)
        return salmap


def rectangle_union(recta, rectb):
    """Calculates the rectangle which results from the union of two rectangles.
    """
    x, y = min(recta[0], rectb[0]), min(recta[1], rectb[1])
    w = max(recta[0] + recta[2], rectb[0] + rectb[2]) - x
    h = max(recta[1] + recta[3], rectb[1] + rectb[3]) - y
    return x, y, w, h


def rectangle_intersection(recta, rectb):
    """Calculates the rectangle which results from the intersection of two rectangles."""
    x, y = max(recta[0], rectb[0]), max(recta[1], rectb[1])
    w = min(recta[0] + recta[2], rectb[0] + rectb[2]) - x
    h = min(recta[1] + recta[3], rectb[1] + rectb[3]) - y
    return (x, y, w, h) if w > 0 and h > 0 else (0, 0, 0, 0)


def rectangle_scale(rect, scale):
    """Calculates the rectangle scaled by a defined scale factor."""
    return [round(rect[0] * scale[0]), round(rect[1] * scale[1]),
            round(rect[2] * scale[0]), round(rect[3] * scale[1])]


def rectangle_zoom(rect, zoom):
    """Applies a zoom factor to the rectangle."""
    zw, zh = rect[2] / zoom, rect[3] / zoom
    dx, dy = (zw - rect[2]) / 2, (zh - rect[3]) / 2
    return [round(rect[0] - dx), round(rect[1] - dy), round(zw), round(zh)]


def laplacian(img):
    """Calculates the laplacian edges for a given image."""
    img = cv2.GaussianBlur(img, (5, 5), 0)
    lap = cv2.Laplacian(img, cv2.CV_32F)
    mag = cv2.convertScaleAbs(lap)
    mag = cv2.GaussianBlur(mag, (3, 3), 0)
    if mag.ndim == 3:
        mag = np.max(mag, axis=2)
    mag[mag < np.mean(mag)] = 0
    return mag


def max_location(salmap):
    smap = salmap.copy()
    thick = int(0.1 * min(salmap.shape))
    x, y, w, h = (0, 0, salmap.shape[1], salmap.shape[0])
    cv2.rectangle(smap, (int(x), int(y)), (int(x+w-1), int(y+h-1)), 0, thick)
    minVal, maxVal, minLoc, maxLoc = cv2.minMaxLoc(smap)
    return maxLoc


def resize_image(img, size, filter=Image.Resampling.LANCZOS):
    im = Image.fromarray(img)
    return np.array(im.resize(size, filter))


def image_roi(img, rect):
    x, y, w, h = rect
    return img[y:y+h, x:x+w]


def crop_region(salmap, faces=[]):
    h, w = salmap.shape
    if len(faces) > 0:
        recti = faces[0]
        for face in faces:
            recti = rectangle_union(recti, face)
    else:
        x, y = max_location(salmap)
        recti = rectangle_zoom((x-1, y-1, 2, 2), 0.2)
    recti = rectangle_zoom(recti, 0.9)
    recto = [0, 0, w, h]
    return recti, recto


def saliency_area(salmap, rect):
    x, y, w, h = rect
    mask = np.zeros(salmap.shape, dtype=bool)
    mask[y:y+h, x:x+w] = True
    return np.sum(salmap[mask])


def content_preservation(salmap, rect):
    sarea = saliency_area(salmap, rect)
    return sarea / np.sum(salmap)


def boundary_simplicity(energy, rect):
    x, y, w, h = rect
    top = energy[y, x:x+w]
    bottom = energy[y+h-1, x:x+w]
    left = energy[y:y+h, x]
    right = energy[y:y+h, x+w-1]
    boundary = np.hstack([top, bottom, left, right])
    # boundary = boundary[boundary > np.mean(boundary)]
    return np.mean(boundary)


def ratio_size(width, ratio):
    return int(width), int(width / ratio)


def maximum_width(size, ratio):
    return int(min(size[0], ratio * min(size[1], size[0] / ratio)))


def sliding_size(ratio, step, width):
    for stp in range(0, int(0.9 * width), step):
        sw, sh = ratio_size(width - stp, ratio)
        if sh <= 0:
            return
        yield (sw, sh)


def sliding_window(img, step, size, rect):
    ww, wh = size
    rx, ry, rw, rh = rect
    x1, y1, x2, y2 = rx, ry, rx+rw, ry+rh
    for y in range(y1, y2-wh+1, step):
        for x in range(x1, x2-ww+1, step):
            yield (x, y, img[y:y + wh, x:x + ww])


def max_pooling(img):
    w = img.shape[1] >> 1 << 1
    h = img.shape[0] >> 1 << 1
    dstack = np.dstack([img[0:h:2, 0:w:2], img[0:h:2, 1:w:2],
                        img[1:h:2, 0:w:2], img[1:h:2, 1:w:2]])
    return np.max(dstack, axis=2)


def attention_windows(salmap, ratio, recti, recto, step=2):
    xi, yi, wi, hi = recti
    m = maximum_width((recto[2], recto[3]), ratio)
    smean = np.mean(salmap)
    areas = []
    maxmean = 0
    for sw, sh in sliding_size(ratio, step, m):
        recta = rectangle_union([xi+wi-sw, yi+hi-sh, sw, sh], [xi, yi, sw, sh])
        rect = rectangle_intersection(recta, recto)
        rects, means = [], []
        for (x, y, window) in sliding_window(salmap, step, (sw, sh), rect):
            mean = np.mean(window)
            if mean > maxmean:
                if not areas or 0.1 < (np.sum(window > smean) / (sw * sh)) < 0.7:  # negative space
                    rects.append([x, y, sw, sh])
                    means.append(mean)
        if len(means) > 0:
            rects = np.array(rects)
            means = np.array(means)
            maxmean = np.max(means)
            idxs = means > 0.9 * maxmean
            rects = rects[idxs]
            means = means[idxs]
            idxs = np.argsort(means)[::-1]
            areas.extend(rects[idxs])
        if sw < wi or sh < hi:
            break
        break # disable crop zoom
    return np.array(areas)


def best_crop_area(salmap, faces, energy, ratio):
    recti, recto = crop_region(salmap, faces)
    salmap = max_pooling(salmap)
    energy = max_pooling(energy)
    energy = energy.astype(np.float32) * salmap
    rectm = [1, 1, salmap.shape[1]-2, salmap.shape[0]-2]
    recto = rectangle_scale(recto, (0.5, 0.5))
    recto = rectangle_intersection(rectm, recto)
    recti = rectangle_scale(recti, (0.5, 0.5))
    recti = rectangle_intersection(recto, recti)
    rects = attention_windows(salmap, ratio, recti, recto)
    preservation = np.array([content_preservation(salmap, rect) for rect in rects])
    simplicities = np.array([boundary_simplicity(energy, rect) for rect in rects])
    content = preservation > 0.8 * np.max(preservation)
    if not np.all(content):
        simplicities = simplicities[:np.argmin(content)]
    idx = np.argmin(simplicities)
    return rectangle_scale(rects[idx], (2, 2))


class Smartcrop:
    def __init__(self):
        self.faces = Faces()
        self.saliency = Saliency()

    def preprocess(self, img):
        height, width = img.shape[:2]
        max_size, min_size = (300, 300), (8, 8)
        max_scale = max(width / max_size[0], height / max_size[1], 1)
        min_scale = min(width / min_size[0], height / min_size[1])
        scale = max_scale if min_scale > max_scale else min_scale
        size = round(width / scale), round(height / scale)
        return cv2.resize(img, size, interpolation=cv2.INTER_AREA)

    def detect(self, img, size):
        h, w = img.shape[:2]
        if (size[0] != w or size[1] != h) and (w > 24 and h > 24):
            rimg = self.preprocess(img)
            rh, rw = rimg.shape[:2]
            energy = laplacian(rimg)
            faces = self.faces.detect(rimg)
            salmap = self.saliency.predict(rimg, faces=faces)
            ar = max(size[0] / size[1], 0.027)
            rect = best_crop_area(salmap, faces, energy, ar)
            return rectangle_scale(rect, (w / rw, h / rh))
        return [0, 0, w, h]

    def transform(self, img, size):
        rect = self.detect(img, size)
        img = image_roi(img, rect)
        return resize_image(img, size)



import numpy as np
from PIL import Image


class Preprocessor:
    """
    Our approach to the CLIP `preprocess` neural net that does not rely on PyTorch.
    The two approaches fully match.
    """

    # Fixed variables that ensure the correct output shapes and values for the `Model` class.
    CLIP_INPUT_SIZE = 224
    # Normalization constants taken from original CLIP:
    # https://github.com/openai/CLIP/blob/3702849800aa56e2223035bccd1c6ef91c704ca8/clip/clip.py#L85
    NORM_MEAN = np.array([0.48145466, 0.4578275, 0.40821073]).reshape((1, 1, 3))
    NORM_STD = np.array([0.26862954, 0.26130258, 0.27577711]).reshape((1, 1, 3))

    @staticmethod
    def _crop_and_resize(img: np.ndarray) -> np.ndarray:
        """Resize and crop an image to a square, preserving the aspect ratio."""

        h, w = img.shape[0:2]
        target_size = Preprocessor.CLIP_INPUT_SIZE

        # Resize so that the smaller dimension matches the required input size.
        resized_h = target_size if h < w else target_size * h // w
        resized_w = target_size * w // h if h < w else target_size

        img_pil = Image.fromarray(img)
        img_pil = img_pil.resize((resized_w, resized_h), resample=Image.BICUBIC)
        img = np.array(img_pil)

        # Now crop to a square
        y_from, x_from = (resized_h - target_size) // 2, (resized_w - target_size) // 2
        img = img[y_from : y_from + target_size, x_from : x_from + target_size, :]
        return img

    def encode_image(self, img: np.ndarray) -> np.ndarray:
        """Preprocesses the images like CLIP's preprocess() function:
        https://github.com/openai/CLIP/blob/3702849800aa56e2223035bccd1c6ef91c704ca8/clip/clip.py#L79

        Args:
            img: numpy array

        Returns:
            img: numpy image after resizing, center cropping and normalization.
        """

        img = Preprocessor._crop_and_resize(img)
        img = img.astype(np.float32) / 255
        img = np.clip(img, 0, 1)  # In case of rounding errors

        # Normalize channels
        img = (img - Preprocessor.NORM_MEAN) / Preprocessor.NORM_STD

        # Mimic the pytorch tensor format for Model class
        img = np.transpose(img, (2, 0, 1))
        img = np.expand_dims(img, 0)
        img = img.astype(np.float32)

        return img



import os
import cv2


class Video:
    def __init__(self, src=0, dest=None):
        self.out = None
        self.quit = False
        self.win_name = ''
        self.cap = cv2.VideoCapture(src)
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)
        self.frames = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.width = int(self.cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        self.height = int(self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        self.duration = round(self.frames / self.fps, 3)
        self.frame_rate = self.fps
        if dest:
            dirname = os.path.dirname(dest)
            if dirname:
                os.makedirs(dirname, exist_ok=True)
            fourcc = cv2.VideoWriter_fourcc(*'avc1')
            self.out = cv2.VideoWriter(dest, fourcc, self.fps, (self.width, self.height))

    def __len__(self):
        return self.frames

    def __iter__(self):
        while self.cap.isOpened():
            ret, frame = self.cap.read()
            if ret is False or frame is None or self.quit:
                break
            yield cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        self.cap.release()
        if self.out:
            self.out.release()
        if self.win_name:
            cv2.destroyWindow(self.win_name)
            cv2.waitKey(1)

    def write(self, frame):
        self.out.write(cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))
    
    def show(self, frame):
        if not self.win_name:
            self.win_name = 'Video'
            cv2.namedWindow(self.win_name, cv2.WINDOW_GUI_NORMAL)
        cv2.imshow(self.win_name, cv2.cvtColor(frame, cv2.COLOR_RGB2BGR))
        ch = 0xFF & cv2.waitKey(int(self.fps))
        if (ch == 27 or ch == ord('q')) or cv2.getWindowProperty(self.win_name, cv2.WND_PROP_VISIBLE) < 1:
            self.quit = True



import os
import json
import base64
import tempfile
import requests
import mimetypes
import numpy as np
import onnxruntime as ort

from tqdm import tqdm
from io import BytesIO
from pathlib import Path
from PIL import Image, ImageOps

from .video import Video
from .sketcher import Sketcher
from .draw import get_color, render_results

tempdir = tempfile.gettempdir()

API_URL = 'https://api.abraia.me'

HEADERS = {'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36'}

mimetypes.add_type('image/webp', '.webp')


def get_type(path):
    return mimetypes.guess_type(path)[0] or 'binary/octet-stream'


class NumpyEncoder(json.JSONEncoder):
    """Special json encoder for numpy types."""
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def is_url(url):
    return url.startswith('http://') or url.startswith('https://')


def url_path(path):
    return f"{API_URL}/files/{path}"


def make_dirs(dest):
    """Create directory if it doesn't exist."""
    dirname = os.path.dirname(dest)
    if dirname:
        os.makedirs(dirname, exist_ok=True)


def temporal_src(path):
    dest = os.path.join(tempdir, path)
    make_dirs(dest)
    return dest


def download_url(url: str, dest: str, chunk_size: int = 8192):
    filename = os.path.basename(dest)
    temp_dest = Path(dest).with_name(filename + '.part')
    temp_dest.parent.mkdir(parents=True, exist_ok=True)
    with requests.get(url, headers=HEADERS, stream=True, allow_redirects=True) as r:
        r.raise_for_status()
        total = int(r.headers.get('content-length', 0))
        with open(temp_dest, 'wb') as f, tqdm(desc=filename, total=total, unit='iB', unit_scale=True, unit_divisor=1024) as bar:
            for chunk in r.iter_content(chunk_size=chunk_size): 
                size = f.write(chunk)
                bar.update(size)
            f.flush()
    temp_dest.rename(dest)


def download_file(path):
    dest = temporal_src(path)
    if not os.path.exists(dest):
        download_url(url_path(path), dest)
    return dest


def load_url(url):
    return requests.get(url, headers=HEADERS, stream=True, allow_redirects=True).raw


def load_json(src):
    with open(src, 'r') as f:
        return json.load(f)
    

def save_json(data, dest):
    make_dirs(dest)
    with open(dest, 'w') as f:
        f.write(json.dumps(data, cls=NumpyEncoder))
    return dest


def load_data(src):
    with open(src, 'rb') as f:
        return f.read()


def save_data(dest, data):
    make_dirs(dest)
    with open(dest, 'wb') as f:
        f.write(data)
    return dest


def load_image(src, mode='RGB'):
    # Fix image orientation based on its EXIF data
    im = ImageOps.exif_transpose(Image.open(src)) 
    return np.array(im.convert(mode))


def save_image(img, dest):
    make_dirs(dest)
    Image.fromarray(img).save(dest)
    return dest


def show_image(img):
    Image.fromarray(img).show()


def image_base64(img, format='jpeg'):
    im = Image.fromarray(img)
    with BytesIO() as buffer:
        im.save(buffer, format=format)
        encoded = base64.b64encode(buffer.getvalue()).decode()
        return f'data:image/{format};base64,{encoded}'


def get_providers():
    available_providers = ort.get_available_providers()
    providers = ["CUDAExecutionProvider", "CoreMLExecutionProvider", "CPUExecutionProvider"]
    return [provider for provider in available_providers if provider in providers]



import sys
import cv2
import numpy as np
from math import log
from io import BytesIO

from PIL import Image
from PIL import ImageCms
from PIL import ImageColor


def _assert_compatible(im1, im2):
    """Raise an error if the shape and dtype do not match."""
    if not im1.dtype == im2.dtype:
        raise ValueError('Input images must have the same dtype.')
    if not im1.shape == im2.shape:
        raise ValueError('Input images must have the same dimensions.')
    return


def _as_floats(im1, im2):
    """Promote im1, im2 to nearest appropriate floating point precision."""
    if im1.dtype != np.float32:
        im1 = im1.astype(np.float32)
    if im2.dtype != np.float32:
        im2 = im2.astype(np.float32)
    return im1 / 255, im2 / 255


def compare_mse(img1, img2):
    """Compute the mean-squared error (MSE) between two images.

    Arguments:
        img1 (ndarray): Image of any dimensionality.
        img2 (ndarray): Image of any dimensionality.

    Returns:
        mse (float): The mean-squared error (MSE) metric.
    """
    _assert_compatible(img1, img2)
    img1, img2 = _as_floats(img1, img2)
    return np.mean(np.square(img1 - img2))


def compare_psnr(img1, img2):
    """Compute the peak signal to noise ratio (PSNR) comparing two images.

    Arguments:
        img1 (ndarray): Image of any dimensionality.
        img2 (ndarray): Image of any dimensionality.

    Returns:
        psnr (float): The peak signal to noise ratio (PSNR) metric.

    .. [1] https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio
    """
    _assert_compatible(img1, img2)
    img1, img2 = _as_floats(img1, img2)
    _mse = compare_mse(img1, img2)
    if _mse == 0:
        return 100
    else:
        return 10 * np.log10(1 / _mse)


def _preprocess_images(img1, img2):
    _assert_compatible(img1, img2)
    if img1.ndim == 3:
        if img1.shape[2] == 4:
            img1 = img1[:, :, :3]
            img2 = img2[:, :, :3]
        img1 = cv2.cvtColor(img1, cv2.COLOR_RGB2GRAY)
        img2 = cv2.cvtColor(img2, cv2.COLOR_RGB2GRAY)
    return _as_floats(img1, img2)


def _downsample(img1, img2):
    img1 = cv2.resize(img1, None, fx=0.5, fy=0.5, interpolation=cv2.INTER_AREA)
    img2 = cv2.resize(img2, None, fx=0.5, fy=0.5, interpolation=cv2.INTER_AREA)
    return img1, img2


def _ssim_map(img1, img2, C1=6.5025, C2=58.5225):
    """Standard SSIM computation."""
    img1_img1 = cv2.pow(img1, 2)
    img2_img2 = cv2.pow(img2, 2)
    img1_img2 = cv2.multiply(img1, img2)
    mu1 = cv2.GaussianBlur(img1, (11, 11), 1.5)
    mu2 = cv2.GaussianBlur(img2, (11, 11), 1.5)
    mu11 = cv2.GaussianBlur(img1_img1, (11, 11), 1.5, dst=img1_img1)
    mu22 = cv2.GaussianBlur(img2_img2, (11, 11), 1.5, dst=img2_img2)
    mu12 = cv2.GaussianBlur(img1_img2, (11, 11), 1.5, dst=img1_img2)
    mu1_mu2 = cv2.multiply(mu1, mu2)
    mu1_mu1 = cv2.pow(mu1, 2, dst=mu1)
    mu2_mu2 = cv2.pow(mu2, 2, dst=mu2)
    b1 = cv2.addWeighted(mu11, 1, mu1_mu1, -1, 0, dst=mu11)
    b2 = cv2.addWeighted(mu22, 1, mu2_mu2, -1, 0, dst=mu22)
    A2 = cv2.addWeighted(mu12, 2, mu1_mu2, -2, C2, dst=mu12)
    A1 = cv2.addWeighted(mu1_mu2, 1, mu1_mu2, 1, C1, dst=mu1_mu2)
    B1 = cv2.addWeighted(mu1_mu1, 1, mu2_mu2, 1, C1, dst=mu1_mu1)
    B2 = cv2.addWeighted(b1, 1, b2, 1, C2, dst=b1)
    A = cv2.multiply(A1, A2, dst=A1)
    B = cv2.multiply(B1, B2, dst=B1)
    ssim_map = cv2.divide(A, B, dst=A)
    return ssim_map


def compare_ssim(img1, img2):
    """Compute the structural similarity index (SSIM) between two images, to
    address image quality comparison by taking texture into account.

    Arguments:
        img1 (ndarray): Image of any dimensionality.
        img2 (ndarray): Image of any dimensionality.

    Returns:
        ssim: (float): The structural similarity index (SSIM).

    .. [1] Z. Wang, A. C. Bovik, H. R. Sheikh and E. P. Simoncelli, "Image
        quality assessment: From error visibility to structural similarity,"
        IEEE Transactions on Image Processing, vol. 13, no. 4, pp. 600-612,
        Apr. 2004.
        https://ece.uwaterloo.ca/~z70wang/publications/ssim.pdf,

    .. [2] Zhou Wang; Bovik, A.C.; ,"Mean squared error: Love it or leave it?
        A new look at Signal Fidelity Measures," Signal Processing Magazine,
        IEEE, vol. 26, no. 1, pp. 98-117, Jan. 2009.

    .. [3] Avanaki, A. N. (2009). Exact global histogram specification
       optimized for structural similarity. Optical Review, 16, 613-621.
       http://arxiv.org/abs/0901.0065,
    """
    img1, img2 = _preprocess_images(img1, img2)
    C1, C2 = 0.01 * 0.01, 0.03 * 0.03
    S = _ssim_map(img1, img2, C1, C2)
    return S.mean()


def compare_mssim(img1, img2):
    """Compute the multi-scale structural similarity index (MS-SSIM) between
    two images.

    Arguments:
        img1 (ndarray): Image of any dimensionality.
        img2 (ndarray): Image of any dimensionality.

    Returns:
        msssim: (float): The multi-scale structural similarity index (MS-SSIM).

    .. [1] Z. Wang, E. P. Simoncelli and A. C. Bovik, "Multi-scale structural
    similarity for image quality assessment," Invited Paper, IEEE Asilomar
    Conference on Signals, Systems and Computers, Nov. 2003
    """
    img1, img2 = _preprocess_images(img1, img2)
    C1, C2 = 0.01 * 0.01, 0.03 * 0.03
    weights = np.array([0.0448, 0.2856, 0.3001, 0.2363, 0.1333])
    ssims = []
    for l in range(len(weights)):
        if l > 0:
            img1, img2 = _downsample(img1, img2)
        if min(img1.shape) < 11:
            break
        ssim_map = _ssim_map(img1, img2, C1, C2)
        ssims.append(cv2.mean(ssim_map)[0])
    return np.sum(np.array(ssims) * weights[:len(ssims)]) / np.sum(weights[:len(ssims)])


def filter_gaussian(img, size, sigma, mode='same'):
    """To avoid edge effects 'valid' mode will ignore filter radius strip
    around edges."""
    out = cv2.GaussianBlur(img, (size, size), sigma)
    if mode == 'valid':
        pad = (size-1) // 2
        return out[pad:-pad, pad:-pad]
    return out


def downsample(img):
    """Implements the MATLAB filter imfilter(im1, ker, 'symmetric', 'same')."""
    ker = np.ones((2, 2)) / 4
    filtered = cv2.filter2D(
        img, -1, ker, anchor=(0, 0), borderType=cv2.BORDER_REFLECT)
    return filtered[::2, ::2]


def adobe_to_srgb(im):
    """Return a new image that is sRGB when it has a ICC profile."""
    try:
        icc = im.info.get('icc_profile')
        if icc:
            srgb = ImageCms.createProfile('sRGB')
            im = ImageCms.profileToProfile(im, BytesIO(icc), srgb)
    except:
        print('PyCMSError: cannot build transform')
    return im


def alpha_to_color(im, background='white'):
    if im.mode == 'RGBA':
        color = ImageColor.getcolor(background, 'RGB')
        back = Image.new('RGB', im.size, color)
        back.paste(im, (0, 0), im)
        return back
    return im


def mode_to_color(im):
    if im.mode == 'L' or im.mode == 'P':
        if 'transparency' in im.info:
            im = im.convert('RGBA')
        else:
            im = im.convert('RGB')
    elif im.mode == 'LA':
        im = im.convert('RGBA')
    elif im.mode == 'RGBX':
        im = im.convert('RGB')
    return im


def convert_mode(im, mode, background='white'):
    im = adobe_to_srgb(im) # color management function
    if im.mode == mode:
        return im
    if mode == 'RGB':
        return alpha_to_color(im, background)
    else:
        im.convert(mode)
        return im


def to_array(im):
    im = mode_to_color(im)
    return np.asarray(im)


def resize_image(img):
    if img.ndim == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    if img.shape[1] > 1200:
        scale = (0.5 * (img.shape[1] - 1200) + 1200) / img.shape[1]
        img = cv2.resize(
            img, None, fx=scale, fy=scale, interpolation=cv2.INTER_AREA)
    return img


def luminance(img):
    limg = cv2.cvtColor(img, cv2.COLOR_RGB2LAB)[:, :, 0]
    return np.mean(limg) / 255


def red_ratio(img):
    kernel = np.ones((8, 8), np.uint8)
    hsv = cv2.cvtColor(cv2.blur(img, (4, 4)), cv2.COLOR_RGB2HSV)
    red1 = cv2.inRange(hsv, (0, 240, 180), (10, 255, 255))
    red2 = cv2.inRange(hsv, (170, 240, 180), (180, 255, 255))
    binary = cv2.bitwise_or(red1, red2)
    return np.mean(cv2.erode(binary, kernel)) / 255


def optimal_threshold(img):
    area = img.shape[0] * img.shape[1]
    # k = max(0.0021 * log(7.5e-6 * area), 0)
    # thr = k * (1 - luminance(img)) + 0.45 * k + 0.0003
    # thr = 0.4 * thr if red_ratio(img) > 0.004 else thr
    k = max(0.00089 * log(8.9e-6 * area), 0)
    thr = 10/3 * k * (1 - luminance(img)) + k + 0.0003
    return thr


def minimal_quality(img):
    area = img.shape[0] * img.shape[1]
    # qmin = 100 - 7.5 * log(0.00004 * area)
    qmin = min(100 - 6.5 * log(0.00006 * area), 95)
    return int(qmin)


def red_correction(img, thr, qmin):
    red = red_ratio(img)
    thr = 0.33 * thr if red > 0.007 else thr
    qmin = (100 + qmin) / 2 if red > 0.05 * 0.007 else qmin
    return thr, int(qmin)


def rgb2pca(img):
    X = img.reshape(-1, img.shape[-1])
    mean, eigenvectors = cv2.PCACompute(X, np.float32([]))
    return cv2.PCAProject(X, mean, eigenvectors).reshape(img.shape)


def saturation(img):
    kernel = np.ones((9, 9), np.uint8)
    hsv = cv2.cvtColor(cv2.blur(img, (4, 4)), cv2.COLOR_RGB2HSV)
    binary = cv2.inRange(hsv, (0, 200, 80), (255, 255, 255))
    return np.mean(cv2.erode(binary, kernel)) / 255


def salient_sat(img):
    imgpca = rgb2pca(cv2.blur(img, (4, 4)))
    L2_ab = np.sqrt(np.square(imgpca[:, :, 1]) + np.square(imgpca[:, :, 2]))
    return np.max(L2_ab)


def subsample(img):
    subsample = (saturation(img) < 0.0042) and (salient_sat(img) < 120)
    return subsample


def number_of_colors(img):
    if img.ndim == 2:
        img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    hist = cv2.calcHist(
        [img], [0, 1, 2], None, [16, 64, 32], [0, 256, 0, 256, 0, 256])
    ncolors = np.sum(hist > 0.0002 * (img.shape[0] * img.shape[1]))
    return ncolors


def getsize(stream):
    if isinstance(stream, BytesIO):
        return sys.getsizeof(stream)
    return len(stream)


def stats(original, optimized):
    orisize = getsize(original) / 1024
    optsize = getsize(optimized) / 1024
    reduction = 100 * (1 - (optsize / orisize))
    return orisize, optsize, reduction


def encode_jpeg(im, quality, subsampling, progressive):
    output = BytesIO()
    im.save(output, 'JPEG', quality=quality, progressive=progressive, subsampling=subsampling, optimize=True)
    output.seek(0)
    return output


def optimal_quality(im, thr=0.003, qmin=60, qmax=95):
    img = to_array(im)
    rimg = resize_image(img)
    rim = Image.fromarray(rimg)
    progressive = True
    thr = optimal_threshold(img)
    qmin = minimal_quality(img)
    thr, qmin = red_correction(img, thr, qmin)
    gray1 = rimg if rimg.ndim == 2 else cv2.cvtColor(rimg, cv2.COLOR_RGB2GRAY)
    gray2 = np.zeros(gray1.shape, dtype=np.uint8)
    q0, q1 = 100, qmin
    d0, d1 = 0, 1
    q = q1
    for k in range(5):
        fileobj = encode_jpeg(rim, q, 0, False)
        progressive = getsize(fileobj) > 10240
        gray2[...] = cv2.cvtColor(to_array(Image.open(fileobj)), cv2.COLOR_RGB2GRAY)
        mssim = compare_mssim(gray1, gray2)
        d = 1 / mssim - 1
        if abs(thr - d) < (0.05 * thr):
            break
        if d > thr and d < d1:
            d1, q1 = d, q
        if d < thr and d > d0:
            d0, q0 = d, q
        q = int(q0 + (thr - d0) * (q1 - q0) / (d1 - d0 + 1e-10))
        if q1 == q or q1 == q0 or q > 100 or q < qmin:
            break
    q = np.clip(q, qmin, qmax)
    subsampling = 2 if subsample(img) else 0
    return int(q), progressive, subsampling
    

def save_jpeg(im, dest, quality=None, subsampling=2, progressive=False):
    im = alpha_to_color(mode_to_color(im))
    if quality is None:
        quality, progressive, subsampling = optimal_quality(im)
    im.save(dest, 'JPEG', quality=quality, progressive=progressive, subsampling=subsampling, optimize=True)
    return dest


def save_png(im, dest, quality=None):
    if quality is None or quality < 100:
        im = mode_to_color(im)
        im = im.quantize(colors=256, method=3)
    im.save(dest, format='PNG', optimize=True)
    return dest


def save_webp(im, dest, quality=75):
    lossless = True if quality == 100 else False
    im.save(dest, format='WEBP', lossless=lossless, quality=quality, optimize=True)
    return dest



from ..client import Abraia

import os
import io
import contextlib
import numpy as np
#os.environ['YOLO_VERBOSE'] = 'False'

from ultralytics import YOLO


abraia = Abraia()


def sorted_folders(dir):
    items = [os.path.join(dir, name) for name in os.listdir(dir)]
    sorted_items = sorted(items, key=os.path.getctime)
    return sorted_items


def build_model_name(model_name, task):
    if task == 'segment':
        model_name = f"{model_name}-seg"
    if task == 'classify':
        model_name = f"{model_name}-cls"
    return model_name


class Model:
    def __init__(self, task, model_type='yolov8n'):
        model_name = build_model_name(model_type, task)
        self.model = YOLO(f"{model_name}.pt", verbose=False)
        self.model_name = model_name
        self.metrics = {}
        self.task = task

    def train(self, dataset, epochs=100, batch=32, imgsz=640):
        data = f"{dataset}" if self.task == 'classify' else f"{dataset}/data.yaml"
        results = self.model.train(data=data, batch=batch, epochs=epochs, imgsz=imgsz)
        # TODO: Merge with test and add parse metrics
        metrics = self.model.val(data=data)
        # self.metrics = self.test(split='val')
        return metrics

    def test(self, split='val'):
        out = io.StringIO()
        with contextlib.redirect_stderr(out):
            metrics = self.model.val(split=split)
        return {'mAP': metrics.box.map50, 'P': metrics.box.p, 'R': metrics.box.r, 
                'confusionMatrix': metrics.confusion_matrix.matrix}

    def save(self, dataset, classes, imgsz=640, device="cpu"):
        # TODO: Add versioning
        model_src = self.model.export(format="onnx", device=device)
        abraia.upload_file(model_src, f"{dataset}/{self.model_name}.onnx")
        abraia.save_json(f"{dataset}/{self.model_name}.json", 
                         {'task': self.task, 'inputShape': [1, 3, imgsz, imgsz], 
                          'classes': classes, 'metrics': self.metrics})

    def run(self, img):
        objects = []
        results = self.model.predict(img, verbose=False)[0]
        if results:
            for k, box in enumerate(results.boxes):
                class_id = int(box.cls)
                label = results.names[class_id]
                score = float(box.conf)
                x1, y1, x2, y2 = box.xyxy.squeeze().tolist()
                x1, y1, x2, y2 = round(x1), round(y1), round(x2), round(y2)
                object = {'label': label, 'score': score, 'box': [x1, y1, x2 - x1, y2 - y1]}
                if self.task == 'segment':
                    object['polygon'] = results.masks[k].xy[0]
                objects.append(object)
        return objects



import cv2
import numpy as np

from .removebg import BackgroundRemover
from .upscale import Upscaler, SwinIR
from .smartcrop import Smartcrop
from .inpaint import LAMA
from .sam import SAM

from ..inference import PlateDetector
from ..inference.faces import FaceRecognizer
from ..utils import draw, Sketcher


def detect_faces(img):
    recognition = FaceRecognizer()
    return recognition.detect_faces(img)


def detect_plates(img):
    plate = PlateDetector()
    return plate.detect(img)


def detect_smartcrop(img, size):
    smartcrop = Smartcrop()
    return smartcrop.detect(img, size)


def build_mask(img, plates, faces):
    mask = np.zeros(img.shape[:2], np.uint8)
    [draw.draw_filled_polygon(mask, result['polygon'], 255) for result in plates]
    [draw.draw_filled_ellipse(mask, result['box'], 255) for result in faces]
    return mask


def anonymize_image(img):
    plates = detect_plates(img)
    faces = detect_faces(img)
    mask = build_mask(img, plates, faces)
    out = draw.draw_blurred_mask(img, mask)
    return out


def remove_background(img):
    removebg = BackgroundRemover()
    out = removebg.remove(img)
    return out


def blur_background(img):
    mask = (np.ones(img.shape[:2]) * 255).astype(np.uint8)
    back = draw.draw_blurred_mask(img.copy(), mask)
    fore = remove_background(img)
    out = draw.draw_overlay(back, fore)
    return out


def upscale_image(img):
    if max(img.shape) > 1920:
        h, w = img.shape[:2]
        scale = 1920 / max(img.shape)
        size = (round(scale * w), round(scale * h))
        img = cv2.resize(img, size, cv2.INTER_LINEAR)
    upscaler = Upscaler()
    out = upscaler.upscale(img)
    return out


def smartcrop_image(img, size):
    smartcrop = Smartcrop()
    return smartcrop.transform(img, size)


def inpaint_image(img, mask):
    lama = LAMA()
    return lama.inpaint(img, mask)


def clean_image(img):
    sam = SAM()
    lama = LAMA()
    sam.encode(img)

    def handle_click(point):
        mask = sam.predict(img, f'[{{"type":"point","data":[{point[0]},{point[1]}],"label":1}}]')
        sketcher.mask = cv2.bitwise_or(sketcher.dilate(mask), sketcher.mask)
        sketcher.show(sketcher.img, sketcher.mask)
        return lama.inpaint(img, sketcher.mask)

    sketcher = Sketcher(img)
    sketcher.on_click(handle_click)
    out = sketcher.run()
    return out


# __all__ = [clean_image]



import cv2
import numpy as np
import onnxruntime as ort

from ..utils import download_file


def tiling(img, tile_size, overlap = 8):
    """Return an image in a tiled manner."""
    tile_width, tile_height = tile_size
    height, width = img.shape[:2]
    for y in range(0, height, tile_height - overlap):
        y_end = min(y + tile_height, height)
        y = max(0, y_end - tile_height)
        for x in range(0, width, tile_width - overlap):
            x_end = min(x + tile_width, width)
            x = max(0, x_end - tile_width)
            tile = img[y:y_end, x:x_end]
            yield tile


def stitching(tiles, img_size, overlap = 8):
    """Return an image from a set of tiles."""
    tile_height, tile_width = tiles[0].shape[:2]
    width, height = img_size
    out = np.empty((height, width, 3), dtype=np.uint8)
    k = 0
    for y in range(0, height, tile_height - overlap):
        for x in range(0, width, tile_width - overlap):
            x_end = min(x + tile_width, width)
            y_end = min(y + tile_height, height)
            x = max(0, x_end - tile_width)
            y = max(0, y_end - tile_height)
            out[y:y_end, x:x_end, :] = tiles[k]
            k += 1
    return out


def ceil_modulo(x, mod):
    return x if x % mod == 0 else (x // mod + 1) * mod


def pad_img_to_modulo(img, mod):
    channels, height, width = img.shape
    out_height, out_width = ceil_modulo(height, mod), ceil_modulo(width, mod)
    return np.pad(img, ((0, 0), (0, out_height - height), (0, out_width - width)), mode="symmetric")


class LAMA:
    
    def __init__(self):
        self.image_size = (512, 512)
        sess_options = ort.SessionOptions()
        model_src = download_file('multiple/models/editing/lama_fp32.onnx')
        self.session = ort.InferenceSession(model_src, sess_options=sess_options)

    def preprocess(self, image, mask, pad_out_to_modulo=8):
        out_image = cv2.resize(image, self.image_size, interpolation=cv2.INTER_LINEAR)
        out_mask = cv2.resize(mask, self.image_size, interpolation=cv2.INTER_NEAREST)

        out_image = (out_image.transpose(2, 0, 1) / 255)
        out_mask = (out_mask[np.newaxis, ...] / 255)

        if pad_out_to_modulo is not None and pad_out_to_modulo > 1:
            out_image = pad_img_to_modulo(out_image, pad_out_to_modulo)
            out_mask = pad_img_to_modulo(out_mask, pad_out_to_modulo)

        out_mask = (out_mask > 0) * 1

        out_image = np.expand_dims(out_image, axis=0).astype(np.float32)
        out_mask = np.expand_dims(out_mask, axis=0).astype(np.float32)
        return out_image, out_mask
    
    def postprocess(self, output, size):
        output = output[0].transpose(1, 2, 0).astype(np.uint8)
        output = cv2.resize(output, size, interpolation=cv2.INTER_CUBIC)
        return output

    def process(self, img, mask):
        h, w = img.shape[:2]
        image, mask = self.preprocess(img, mask)
        outputs = self.session.run(None, {'image': image, 'mask': mask})
        output = self.postprocess(outputs[0], (w, h))
        return output
    
    def inpaint(self, img, mask):
        outs = []
        tile_size = self.image_size
        img_size = (img.shape[1], img.shape[0])
        for im, msk in zip(tiling(img, tile_size), tiling(mask, tile_size)):
            outs.append(self.process(im, msk))
        out = stitching(outs, img_size)
        return out
    



from ..client import Abraia
from ..utils import url_path, make_dirs
from . import classify, detect

import os
import shutil
import itertools
from PIL import Image
from tqdm.contrib.concurrent import process_map
from sklearn.model_selection import train_test_split


abraia = Abraia()


def list_datasets():
    folders = abraia.list_files()[1]
    return [folder['name'] for folder in folders if abraia.check_file(f"{folder['name']}/annotations.json")]


def list_images(project):
    files = abraia.list_files(f"{project}/")[0]
    dataset = [f for f in files if f['type'].startswith('image/')]
    for data in dataset:
        data['url'] = url_path(f"{abraia.userid}/{data['path']}")
    return dataset


def load_annotations(project):
    annotations = abraia.load_json(f"{project}/annotations.json")
    for annotation in annotations:
        annotation['path'] = f"{project}/{annotation['filename']}"
    return annotations


def save_annotations(project, annotations):
    abraia.save_json(f"{project}/annotations.json", annotations)


def load_labels(annotations):
    labels = []
    for annotation in annotations:
        for object in annotation.get('objects', []):
            label = object.get('label')
            if label and label not in labels:
                labels.append(label)
    return list(set(labels))


def load_task(annotations):
    classify, detect, segment = False, False, False
    for annotation in annotations:
        for object in annotation.get('objects', []):
            if 'polygon' in object:
                segment = True
            elif 'box' in object:
                detect = True
            elif 'label' in object:
                classify = True
    if segment:
        return 'segment'
    if detect:
        return 'detect'
    if classify:
        return 'classify'


def load_tasks(annotations):
    tasks = ['classify', 'detect', 'segment']
    task = load_task(annotations)
    if task:
        idx = tasks.index(task)
        return tasks[:idx+1]
    return []


def download_file(path, folder):
    dest = os.path.join(folder, os.path.basename(path))
    if not os.path.exists(dest):
        abraia.download_file(path, dest)
    return dest


def save_annotation(annotation, folder, classes, task):
    if task == 'classify':
        for object in annotation.get('objects', []):
            label = object.get('label')
            if label:
                src = os.path.join(folder, annotation['filename'])
                dest = os.path.join(folder, label, annotation['filename'])
                shutil.move(src, dest)
                break
    else:
        im = Image.open(os.path.join(folder, 'images', annotation['filename']))
        label_lines = []
        for object in annotation.get('objects', []):
            label, box, polygon = object.get('label'), object.get('box'), object.get('polygon')
            # Convert polygon or box to yolo format
            if task == 'segment':
                if polygon:
                    label_line = f"{classes.index(label)} " + ' '.join([f"{point[0] / im.width} {point[1] / im.height}" for point in polygon])
                    label_lines.append(label_line)
            elif task == 'detect':
                if polygon:
                    xx, yy = [point[0] for point in polygon], [point[1] for point in polygon]
                    x1, y1, x2, y2 = min(xx), min(yy), max(xx), max(yy)
                    box = [x1, y1, x2 - x1, y2 - y1]
                if box:
                    label_line = f"{classes.index(label)} {(box[0] + box[2] / 2) / im.width} {(box[1] + box[3] / 2) / im.height} {box[2] / im.width} {box[3] / im.height}"
                    label_lines.append(label_line)
        label_path = os.path.join(folder, 'labels',  f"{os.path.splitext(annotation['filename'])[0]}.txt")
        with open(label_path, 'w') as f:
            f.write('\n'.join(label_lines))


def save_data(annotations, folder, classes, task):
    if (task == 'classify'):
        make_dirs(os.path.join(folder, ''))
        paths = [annotation['path'] for annotation in annotations]
        process_map(download_file, paths, itertools.repeat(folder), max_workers=5)
        for label in classes:
            make_dirs(os.path.join(folder, label, ''))
        for annotation in annotations:
            save_annotation(annotation, folder, classes, task)
    else:
        make_dirs(os.path.join(folder, 'images', ''))
        paths = [annotation['path'] for annotation in annotations]
        process_map(download_file, paths, itertools.repeat(os.path.join(folder, 'images')), max_workers=5)
        make_dirs(os.path.join(folder, 'labels', ''))
        for annotation in annotations:
            save_annotation(annotation, folder, classes, task)


def save_config(dataset, classes):
    yaml_content = f'''
    train: {os.path.join(os.getcwd(), dataset, 'train/images')}
    val: {os.path.join(os.getcwd(), dataset, 'val/images')}
    test: {os.path.join(os.getcwd(), dataset, 'test/images')}
    names: {classes}
    '''
    path = os.path.join(dataset, 'data.yaml')
    with open(path, 'w') as f:
        f.write(yaml_content)


def split_dataset(annotations):
    # TODO: Split dataset by classes to avoid class imbalance
    backgrounds = [annotation for annotation in annotations if not annotation.get('objects')]
    annotations = [annotation for annotation in annotations if annotation.get('objects')]
    train, test = train_test_split(annotations, test_size=0.3)
    val, test = train_test_split(test, test_size=0.5)
    train.extend(backgrounds)
    return train, val, test


def create_dataset(dataset, task, classes):
    if not os.path.exists(dataset):
        annotations = load_annotations(dataset)
        train, val, test = split_dataset(annotations)
        data_annotations = {'train': train, 'val': val, 'test': test}
        #TODO: Download files in one single step
        for x in ['train', 'val', 'test']:
            save_data(data_annotations[x], f"{dataset}/{x}", classes, task)
        save_config(dataset, classes)




from abraia.inference.ops import count_objects, segments_intersect, point_in_polygon


class LineCounter:
    def __init__(self, line):
        self.line = line
        self.in_count = 0
        self.out_count = 0
        self.prev_results = {}

    def update(self, results):
        for result in results:
            track_id = result.get('track_id')
            if track_id is not None:
                prev_result = self.prev_results.get(track_id)
                if prev_result:
                    (x1, y1, w1, h1), (x2, y2, w2, h2) = result['box'], prev_result['box']
                    val = segments_intersect(self.line[0], self.line[1], [x1 + w1 / 2, y1 + h1 / 2], [x2 + w2 / 2, y2 + h2 / 2])
                    if val > 0:
                        self.in_count += 1
                    if val < 0:
                        self.out_count += 1
                self.prev_results[track_id] = result
        return self.in_count, self.out_count


class RegionFilter:
    def __init__(self, polygon):
        self.region = polygon

    def update(self, results):
        in_objects, out_objects = [], []
        for result in results:
            box = result.get('box')
            if box:
                x, y, w, h = box
                center = x + w / 2, y + h / 2
                if point_in_polygon(center, self.region):
                    in_objects.append(result)
                else:
                    out_objects.append(result)
        return in_objects, out_objects


class RegionTimer():
    def __init__(self, polygon):
        self.region = polygon
        self.timers = {}

    def update(self, results, frame_time):
        in_objects, out_objects = [], []
        for result in results:
            box = result.get('box')
            track_id = result.get('track_id')
            start_time = self.timers.get(track_id)
            if box and track_id is not None:
                x, y, w, h = box
                center = x + w / 2, y + h / 2
                if point_in_polygon(center, self.region):
                    in_objects.append(result)
                    if start_time == None:
                        self.timers[track_id] = frame_time
                    else:
                        print(f'Time in zone [{track_id}]: {round(frame_time - start_time, 2)}')
                        result['label'] = f"waiting {round(frame_time - start_time, 2)}s"
                        del result['score']
                else:
                    out_objects.append(result)
                    if start_time != None:
                        del self.timers[track_id]
        return in_objects, out_objects



from .detect import Model
from .tracker import Tracker
from .faces import FaceRecognizer, FaceAttribute
from .plates import PlateDetector, PlateRecognizer
from .ocr import TextSystem
    

__all__ = [Model, Tracker, FaceRecognizer, FaceAttribute, PlateDetector, PlateRecognizer, TextSystem]



import os
import gzip
import html
import ftfy
import numpy as np
import regex as re
from typing import Dict, List, Set, Union, Iterable


def default_bpe():
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L10
    BPE is the byte pair encoding that is used within the tokenizer. More can be read here:
    https://huggingface.co/course/chapter6/5?fw=pt
    Returns:
        the vocabulary byte-pair encodings
    """
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), "bpe_simple_vocab_16e6.txt.gz")


def bytes_to_unicode() -> Dict[int, str]:
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L16
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = (list(range(ord("!"), ord("~") + 1))
        + list(range(ord("¡"), ord("¬") + 1))
        + list(range(ord("®"), ord("ÿ") + 1)))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8 + n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))


def get_pairs(word) -> Set[str]:
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L38
    Return set of symbol pairs in a word.
    Word is represented as tuple of symbols (symbols being variable-length strings).
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


def basic_clean(text) -> str:
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L50
    """
    text = ftfy.fix_text(text)
    text = html.unescape(html.unescape(text))
    return text.strip()


def whitespace_clean(text) -> str:
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L56
    """
    text = re.sub(r"\s+", " ", text)
    text = text.strip()
    return text


class Tokenizer(object):
    """Taken from CLIP.
    https://github.com/openai/CLIP/blob/main/clip/simple_tokenizer.py#L62
    """

    def __init__(self, bpe_path: str = default_bpe()):
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        merges = gzip.open(bpe_path).read().decode("utf-8").split("\n")
        merges = merges[1 : 49152 - 256 - 2 + 1]
        merges = [tuple(merge.split()) for merge in merges]
        vocab = list(bytes_to_unicode().values())
        vocab = vocab + [v + "</w>" for v in vocab]
        for merge in merges:
            vocab.append("".join(merge))
        vocab.extend(["<|startoftext|>", "<|endoftext|>"])
        self.encoder = dict(zip(vocab, range(len(vocab))))
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.bpe_ranks = dict(zip(merges, range(len(merges))))
        self.cache = {"<|startoftext|>": "<|startoftext|>", "<|endoftext|>": "<|endoftext|>"}
        self.pat = re.compile(
            r"""<\|startoftext\|>|<\|endoftext\|>|'s|'t|'re|'ve|'m|'ll|'d|[\p{L}]+|[\p{N}]|[^\s\p{L}\p{N}]+""",
            re.IGNORECASE)

    def bpe(self, token: str) -> str:
        if token in self.cache:
            return self.cache[token]
        word = tuple(token[:-1]) + (token[-1] + "</w>",)
        pairs = get_pairs(word)

        if not pairs:
            return token + "</w>"

        while True:
            bigram = min(pairs, key=lambda pair: self.bpe_ranks.get(pair, float("inf")))
            if bigram not in self.bpe_ranks:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except:
                    new_word.extend(word[i:])
                    break
                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = get_pairs(word)
        word = " ".join(word)
        self.cache[token] = word
        return word

    def encode(self, text: str) -> List[int]:
        bpe_tokens = []
        text = whitespace_clean(basic_clean(text)).lower()
        for token in re.findall(self.pat, text):
            token = "".join(self.byte_encoder[b] for b in token.encode("utf-8"))
            bpe_tokens.extend(self.encoder[bpe_token] for bpe_token in self.bpe(token).split(" "))
        return bpe_tokens

    def decode(self, tokens: List[int]) -> str:
        text = "".join([self.decoder[token] for token in tokens])
        text = (bytearray([self.byte_decoder[c] for c in text])
            .decode("utf-8", errors="replace")
            .replace("</w>", " "))
        return text

    def encode_text(self, texts: Union[str, Iterable[str]], context_length: int = 77,
        truncate: bool = False) -> np.array:
        """Taken from CLIP and reformatted to replace pytorch zeros with numpy zeros.
        Furthermore, this has been wrapped inside the Tokenizer class instead of being
        a separate function.
        https://github.com/openai/CLIP/blob/main/clip/clip.py#L197
        Returns the tokenized representation of given input string(s)
        Parameters
        ----------
        texts : Union[str, List[str]]
            An input string or a list of input strings to tokenize
        context_length : int
            The context length to use; all CLIP models use 77 as the context length
        truncate: bool
            Whether to truncate the text in case its encoding is longer than the context length
        Returns
        -------
        A two-dimensional tensor containing the resulting tokens, shape = [number of input strings, context_length].
        """
        if isinstance(texts, str):
            texts = [texts]

        sot_token = self.encoder["<|startoftext|>"]
        eot_token = self.encoder["<|endoftext|>"]
        all_tokens = [[sot_token] + self.encode(text) + [eot_token] for text in texts]
        result = np.zeros((len(all_tokens), context_length), dtype=np.int32)

        for i, tokens in enumerate(all_tokens):
            if len(tokens) > context_length:
                if truncate:
                    tokens = tokens[:context_length]
                    tokens[-1] = eot_token
                else:
                    raise RuntimeError(f"Input {texts[i]} is too long for context length {context_length}")
            result[i, : len(tokens)] = np.array(tokens)

        return result



import os
import cv2
import math
import numpy as np
import onnxruntime as ort

from .ops import non_maximum_suppression, normalize, mask_to_polygon, softmax
from ..utils import download_file, load_json, get_color, get_providers


def resize(img, size):
    scale = max(size / img.shape[1], size / img.shape[0])
    width, height = round(scale * img.shape[1]), round(scale * img.shape[0])
    return cv2.resize(img, (width, height), interpolation=cv2.INTER_LINEAR)


def preprocess(img, size = 224):
    img = resize(img, size)
    img = normalize(img, [0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    return np.expand_dims(img.transpose((2, 0, 1)), axis=0)


def postprocess(outputs, classes):
    probs = softmax(outputs[0].flatten())
    idx = np.argmax(probs)
    return [{'label': classes[idx], 'score': probs[idx], 'color': get_color(idx)}]


def scale_size(size, new_size):
    scale = min(new_size[0] / size[0], new_size[1] / size[1])
    return round(scale * size[0]), round(scale * size[1])


def prepare_input(img, shape):
    """Converts the input image array to a (3, height, width) tensor."""
    size = scale_size((img.shape[1], img.shape[0]), (shape[3], shape[2]))
    img = cv2.resize(img, size, interpolation=cv2.INTER_LINEAR)
    # dw, dh = (shape[3] - size[0]) / 2, (shape[2] - size[1]) / 2
    # left, top, right, bottom = round(dw - 0.1), round(dh - 0.1), round(dw + 0.1), round(dh + 0.1)
    left, top, right, bottom = 0, 0, shape[3] - size[0], shape[2] - size[1]
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114))
    img = img.transpose((2, 0, 1)).reshape(shape).astype(np.float32)
    return img / 255


def sigmoid_mask(z):
    mask = 1 / (1 + np.exp(-z))
    return (255 * mask).astype('uint8')


def get_mask(row, box, size):
    """Extracts the segmentation mask for an object (box) in a row."""
    shape = round(math.sqrt(row.shape[0]))
    mask = row.reshape(shape, shape)
    mask = sigmoid_mask(mask)
    x, y, w, h = box
    mask_x1, mask_y1 = round(x / size[0] * shape), round(y / size[1] * shape)
    mask_x2, mask_y2 = round((x + w) / size[0] * shape), round((y + h) / size[1] * shape)
    mask = mask[mask_y1:mask_y2, mask_x1:mask_x2]
    mask = cv2.resize(mask, (round(w), round(h)), cv2.INTER_LINEAR)
    return mask


def process_output(outputs, size, shape, classes, conf_threshold=0.25, iou_threshold=0.7, approx=0.001):
    """Converts the RAW model output from YOLOv8 to an array of detected
    objects, containing the bounding box, label and the probability.
    """
    output0 = outputs[0][0].astype("float")
    output0 = output0.transpose()
    if len(outputs) == 2:
        output1 = outputs[1][0].astype("float")
        output1 = output1.reshape(output1.shape[0], output1.shape[1] * output1.shape[2])
    img_width, img_height = size
    model_width, model_height = shape[3], shape[2]
    scale = 1 / min(model_width / img_width, model_height / img_height)
    objects = []
    for row in output0:
        xc, yc, w, h = row[:4]
        probs = row[4:4+len(classes)]
        idx = probs.argmax()
        if probs[idx] < conf_threshold:
            continue
        x1, y1 = round((xc - w/2) * scale), round((yc - h/2) * scale)
        x2, y2 = round((xc + w/2) * scale), round((yc + h/2) * scale)
        obj = {'label': classes[idx], 'score': float(probs[idx]), 'box': [x1, y1, x2 - x1, y2 - y1], 'color': get_color(idx)}
        if len(outputs) == 2:
            obj['mask'] = row[4+len(classes):]
        objects.append(obj)
    results = non_maximum_suppression(objects, iou_threshold)
    for result in results:
        if len(outputs) == 2:
            x, y, w, h = result['box']
            mask = result['mask'] @ output1
            size = (round(model_width * scale), round(model_height * scale))
            mask = get_mask(mask, result['box'], size)
            mask = cv2.threshold(mask, 127, 255, cv2.THRESH_BINARY)[1]
            result['polygon'] = mask_to_polygon(mask, (x, y), approx)
            result.pop('mask', None)
            # result['mask'] = mask
    return results


class Model:
    def __init__(self, model_uri):
        config_uri = f"{os.path.splitext(model_uri)[0]}.json"
        self.config = load_json(download_file(config_uri))
        self.session = ort.InferenceSession(download_file(model_uri), providers=get_providers())
        self.input_name = self.session.get_inputs()[0].name
        self.input_shape = self.config['inputShape']

    def run(self, img, conf_threshold=0.35, iou_threshold=0.7, approx=0.001):
        if self.config.get('task'):
            img_size = img.shape[1], img.shape[0]
            inputs = {self.input_name: prepare_input(img, self.input_shape)}
            outputs = self.session.run(None, inputs)
            return process_output(outputs, img_size, self.input_shape, self.config['classes'], conf_threshold, iou_threshold, approx)
        outputs = self.session.run(None, {self.input_name: preprocess(img)})
        return postprocess(outputs, self.config['classes'])



import lap
import numpy as np


class KalmanFilter(object):
    """
    A simple Kalman filter for tracking bounding boxes in image space.

    The 8-dimensional state space

        x, y, a, h, vx, vy, va, vh

    contains the bounding box center position (x, y), aspect ratio a, height h,
    and their respective velocities.

    Object motion follows a constant velocity model. The bounding box location
    (x, y, a, h) is taken as direct observation of the state space (linear
    observation model).

    """

    def __init__(self):
        ndim, dt = 4, 1.

        # Create Kalman filter model matrices.
        self._motion_mat = np.eye(2 * ndim, 2 * ndim)
        for i in range(ndim):
            self._motion_mat[i, ndim + i] = dt
        self._update_mat = np.eye(ndim, 2 * ndim)

        # Motion and observation uncertainty are chosen relative to the current
        # state estimate. These weights control the amount of uncertainty in
        # the model. This is a bit hacky.
        self._std_weight_position = 1. / 20
        self._std_weight_velocity = 1. / 160

    def initiate(self, measurement):
        """Create track from unassociated measurement.

        Parameters
        ----------
        measurement : ndarray
            Bounding box coordinates (x, y, a, h) with center position (x, y),
            aspect ratio a, and height h.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector (8 dimensional) and covariance matrix (8x8
            dimensional) of the new track. Unobserved velocities are initialized
            to 0 mean.

        """
        mean_pos = measurement
        mean_vel = np.zeros_like(mean_pos)
        mean = np.r_[mean_pos, mean_vel]

        std = [
            2 * self._std_weight_position * measurement[3],
            2 * self._std_weight_position * measurement[3],
            1e-2,
            2 * self._std_weight_position * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            10 * self._std_weight_velocity * measurement[3],
            1e-5,
            10 * self._std_weight_velocity * measurement[3]]
        covariance = np.diag(np.square(std))
        return mean, covariance

    def predict(self, mean, covariance):
        """Run Kalman filter prediction step.

        Parameters
        ----------
        mean : ndarray
            The 8 dimensional mean vector of the object state at the previous
            time step.
        covariance : ndarray
            The 8x8 dimensional covariance matrix of the object state at the
            previous time step.

        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.

        """
        std_pos = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-2,
            self._std_weight_position * mean[3]]
        std_vel = [
            self._std_weight_velocity * mean[3],
            self._std_weight_velocity * mean[3],
            1e-5,
            self._std_weight_velocity * mean[3]]
        motion_cov = np.diag(np.square(np.r_[std_pos, std_vel]))

        #mean = np.dot(self._motion_mat, mean)
        mean = np.dot(mean, self._motion_mat.T)
        covariance = np.linalg.multi_dot((
            self._motion_mat, covariance, self._motion_mat.T)) + motion_cov

        return mean, covariance

    def project(self, mean, covariance):
        """Project state distribution to measurement space.

        Parameters
        ----------
        mean : ndarray
            The state's mean vector (8 dimensional array).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).

        Returns
        -------
        (ndarray, ndarray)
            Returns the projected mean and covariance matrix of the given state
            estimate.

        """
        std = [
            self._std_weight_position * mean[3],
            self._std_weight_position * mean[3],
            1e-1,
            self._std_weight_position * mean[3]]
        innovation_cov = np.diag(np.square(std))

        mean = np.dot(self._update_mat, mean)
        covariance = np.linalg.multi_dot((
            self._update_mat, covariance, self._update_mat.T))
        return mean, covariance + innovation_cov

    def multi_predict(self, mean, covariance):
        """Run Kalman filter prediction step (Vectorized version).
        Parameters
        ----------
        mean : ndarray
            The Nx8 dimensional mean matrix of the object states at the previous
            time step.
        covariance : ndarray
            The Nx8x8 dimensional covariance matrics of the object states at the
            previous time step.
        Returns
        -------
        (ndarray, ndarray)
            Returns the mean vector and covariance matrix of the predicted
            state. Unobserved velocities are initialized to 0 mean.
        """
        std_pos = [
            self._std_weight_position * mean[:, 3],
            self._std_weight_position * mean[:, 3],
            1e-2 * np.ones_like(mean[:, 3]),
            self._std_weight_position * mean[:, 3]]
        std_vel = [
            self._std_weight_velocity * mean[:, 3],
            self._std_weight_velocity * mean[:, 3],
            1e-5 * np.ones_like(mean[:, 3]),
            self._std_weight_velocity * mean[:, 3]]
        sqr = np.square(np.r_[std_pos, std_vel]).T

        motion_cov = []
        for i in range(len(mean)):
            motion_cov.append(np.diag(sqr[i]))
        motion_cov = np.asarray(motion_cov)

        mean = np.dot(mean, self._motion_mat.T)
        left = np.dot(self._motion_mat, covariance).transpose((1, 0, 2))
        covariance = np.dot(left, self._motion_mat.T) + motion_cov

        return mean, covariance

    def update(self, mean, covariance, measurement):
        """Run Kalman filter correction step.

        Parameters
        ----------
        mean : ndarray
            The predicted state's mean vector (8 dimensional).
        covariance : ndarray
            The state's covariance matrix (8x8 dimensional).
        measurement : ndarray
            The 4 dimensional measurement vector (x, y, a, h), where (x, y)
            is the center position, a the aspect ratio, and h the height of the
            bounding box.

        Returns
        -------
        (ndarray, ndarray)
            Returns the measurement-corrected state distribution.

        """
        projected_mean, projected_cov = self.project(mean, covariance)
        # chol_factor, lower = scipy.linalg.cho_factor(
        #     projected_cov, lower=True, check_finite=False)
        # kalman_gain = scipy.linalg.cho_solve(
        #     (chol_factor, lower), np.dot(covariance, self._update_mat.T).T,
        #     check_finite=False).T
        kalman_gain = np.linalg.solve(projected_cov, np.dot(covariance, self._update_mat.T).T).T
        innovation = measurement - projected_mean
        new_mean = mean + np.dot(innovation, kalman_gain.T)
        new_covariance = covariance - np.linalg.multi_dot((
            kalman_gain, projected_cov, kalman_gain.T))
        return new_mean, new_covariance


# def linear_assignment(cost_matrix, thresh):
#     if cost_matrix.size == 0:
#         return np.empty((0, 2), dtype=int), tuple(range(cost_matrix.shape[0])), tuple(range(cost_matrix.shape[1]))
#     # Use scipy.optimize.linear_sum_assignment
#     # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html
#     x, y = scipy.optimize.linear_sum_assignment(cost_matrix)  # row x, col y
#     matches = np.asarray([[x[i], y[i]] for i in range(len(x)) if cost_matrix[x[i], y[i]] <= thresh])
#     if len(matches) == 0:
#         unmatched_a = list(np.arange(cost_matrix.shape[0]))
#         unmatched_b = list(np.arange(cost_matrix.shape[1]))
#     else:
#         unmatched_a = list(set(np.arange(cost_matrix.shape[0])) - set(matches[:, 0]))
#         unmatched_b = list(set(np.arange(cost_matrix.shape[1])) - set(matches[:, 1]))
#     return matches, unmatched_a, unmatched_b


def linear_assignment(cost_matrix, thresh):
    if cost_matrix.size == 0:
        return np.empty((0, 2), dtype=int), tuple(range(cost_matrix.shape[0])), tuple(range(cost_matrix.shape[1]))
    matches, unmatched_a, unmatched_b = [], [], []
    cost, x, y = lap.lapjv(cost_matrix, extend_cost=True, cost_limit=thresh)
    for ix, mx in enumerate(x):
        if mx >= 0:
            matches.append([ix, mx])
    unmatched_a = np.where(x < 0)[0]
    unmatched_b = np.where(y < 0)[0]
    matches = np.asarray(matches)
    return matches, unmatched_a, unmatched_b


def box_iou_batch(boxes_true, boxes_detection):
    """
    Compute Intersection over Union (IoU) of two sets of bounding boxes -
        `boxes_true` and `boxes_detection`. Both sets
        of boxes are expected to be in `(x_min, y_min, x_max, y_max)` format.

    Args:
        boxes_true (np.ndarray): 2D `np.ndarray` representing ground-truth boxes.
            `shape = (N, 4)` where `N` is number of true objects.
        boxes_detection (np.ndarray): 2D `np.ndarray` representing detection boxes.
            `shape = (M, 4)` where `M` is number of detected objects.

    Returns:
        np.ndarray: Pairwise IoU of boxes from `boxes_true` and `boxes_detection`.
            `shape = (N, M)` where `N` is number of true objects and
            `M` is number of detected objects.
    """

    def box_area(box):
        return (box[2] - box[0]) * (box[3] - box[1])

    area_true = box_area(boxes_true.T)
    area_detection = box_area(boxes_detection.T)

    top_left = np.maximum(boxes_true[:, None, :2], boxes_detection[:, :2])
    bottom_right = np.minimum(boxes_true[:, None, 2:], boxes_detection[:, 2:])

    area_inter = np.prod(np.clip(bottom_right - top_left, a_min=0, a_max=None), 2)
    ious = area_inter / (area_true[:, None] + area_detection - area_inter)
    ious = np.nan_to_num(ious)
    return ious


def iou_distance(atracks, btracks):
    if (len(atracks) > 0 and isinstance(atracks[0], np.ndarray)) or (
        len(btracks) > 0 and isinstance(btracks[0], np.ndarray)):
        atlbrs = atracks
        btlbrs = btracks
    else:
        atlbrs = [track.tlbr for track in atracks]
        btlbrs = [track.tlbr for track in btracks]
    _ious = np.zeros((len(atlbrs), len(btlbrs)), dtype=np.float32)
    if _ious.size != 0:
        _ious = box_iou_batch(np.asarray(atlbrs), np.asarray(btlbrs))
    cost_matrix = 1 - _ious
    return cost_matrix


def fuse_score(cost_matrix, detections):
    if cost_matrix.size == 0:
        return cost_matrix
    iou_sim = 1 - cost_matrix
    det_scores = np.array([det.score for det in detections])
    det_scores = np.expand_dims(det_scores, axis=0).repeat(cost_matrix.shape[0], axis=0)
    fuse_sim = iou_sim * det_scores
    fuse_cost = 1 - fuse_sim
    return fuse_cost


class TrackState(object):
    New = 0
    Tracked = 1
    Lost = 2
    Removed = 3


class BaseTrack(object):
    _count = 0

    track_id = 0
    is_activated = False
    state = TrackState.New

    features = []
    curr_feature = None
    score = 0
    start_frame = 0
    frame_id = 0
    time_since_update = 0

    @property
    def end_frame(self):
        return self.frame_id

    @staticmethod
    def next_id():
        BaseTrack._count += 1
        return BaseTrack._count

    def activate(self, *args):
        raise NotImplementedError

    def predict(self):
        raise NotImplementedError

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def mark_lost(self):
        self.state = TrackState.Lost

    def mark_removed(self):
        self.state = TrackState.Removed


class STrack(BaseTrack):
    shared_kalman = KalmanFilter()
    def __init__(self, tlwh, score):

        # wait activate
        self._tlwh = np.asarray(tlwh, dtype=np.float32)
        self.kalman_filter = None
        self.mean, self.covariance = None, None
        self.is_activated = False

        self.score = score
        self.tracklet_len = 0

    def predict(self):
        mean_state = self.mean.copy()
        if self.state != TrackState.Tracked:
            mean_state[7] = 0
        self.mean, self.covariance = self.kalman_filter.predict(mean_state, self.covariance)

    @staticmethod
    def multi_predict(stracks):
        if len(stracks) > 0:
            multi_mean = np.asarray([st.mean.copy() for st in stracks])
            multi_covariance = np.asarray([st.covariance for st in stracks])
            for i, st in enumerate(stracks):
                if st.state != TrackState.Tracked:
                    multi_mean[i][7] = 0
            multi_mean, multi_covariance = STrack.shared_kalman.multi_predict(multi_mean, multi_covariance)
            for i, (mean, cov) in enumerate(zip(multi_mean, multi_covariance)):
                stracks[i].mean = mean
                stracks[i].covariance = cov

    def activate(self, kalman_filter, frame_id):
        """Start a new tracklet"""
        self.kalman_filter = kalman_filter
        self.track_id = self.next_id()
        self.mean, self.covariance = self.kalman_filter.initiate(self.tlwh_to_xyah(self._tlwh))

        self.tracklet_len = 0
        self.state = TrackState.Tracked
        if frame_id == 1:
            self.is_activated = True
        # self.is_activated = True
        self.frame_id = frame_id
        self.start_frame = frame_id

    def re_activate(self, new_track, frame_id, new_id=False):
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_track.tlwh)
        )
        self.tracklet_len = 0
        self.state = TrackState.Tracked
        self.is_activated = True
        self.frame_id = frame_id
        if new_id:
            self.track_id = self.next_id()
        self.score = new_track.score

    def update(self, new_track, frame_id):
        """
        Update a matched track
        :type new_track: STrack
        :type frame_id: int
        :type update_feature: bool
        :return:
        """
        self.frame_id = frame_id
        self.tracklet_len += 1

        new_tlwh = new_track.tlwh
        self.mean, self.covariance = self.kalman_filter.update(
            self.mean, self.covariance, self.tlwh_to_xyah(new_tlwh))
        self.state = TrackState.Tracked
        self.is_activated = True

        self.score = new_track.score

    @property
    # @jit(nopython=True)
    def tlwh(self):
        """Get current position in bounding box format `(top left x, top left y,
                width, height)`.
        """
        if self.mean is None:
            return self._tlwh.copy()
        ret = self.mean[:4].copy()
        ret[2] *= ret[3]
        ret[:2] -= ret[2:] / 2
        return ret

    @property
    # @jit(nopython=True)
    def tlbr(self):
        """Convert bounding box to format `(min x, min y, max x, max y)`, i.e.,
        `(top left, bottom right)`.
        """
        ret = self.tlwh.copy()
        ret[2:] += ret[:2]
        return ret

    @staticmethod
    # @jit(nopython=True)
    def tlwh_to_xyah(tlwh):
        """Convert bounding box to format `(center x, center y, aspect ratio,
        height)`, where the aspect ratio is `width / height`.
        """
        ret = np.asarray(tlwh).copy()
        ret[:2] += ret[2:] / 2
        ret[2] /= ret[3]
        return ret

    def to_xyah(self):
        return self.tlwh_to_xyah(self.tlwh)

    @staticmethod
    # @jit(nopython=True)
    def tlbr_to_tlwh(tlbr):
        ret = np.asarray(tlbr).copy()
        ret[2:] -= ret[:2]
        return ret

    @staticmethod
    # @jit(nopython=True)
    def tlwh_to_tlbr(tlwh):
        ret = np.asarray(tlwh).copy()
        ret[2:] += ret[:2]
        return ret

    def __repr__(self):
        return 'OT_{}_({}-{})'.format(self.track_id, self.start_frame, self.end_frame)


class ByteTracker():
    def __init__(self, track_thresh=0.25, track_buffer=30, match_thresh=0.8, frame_rate=30):
        self.tracked_stracks = []
        self.lost_stracks = []
        self.removed_stracks = []

        self.mot20 = False
        self.frame_id = 0
        self.track_thresh = track_thresh
        self.match_thresh = match_thresh
        self.det_thresh = track_thresh + 0.1
        self.buffer_size = int(frame_rate / 30.0 * track_buffer)
        self.max_time_lost = self.buffer_size
        self.kalman_filter = KalmanFilter()

    def update(self, bboxes, scores):
        self.frame_id += 1
        activated_starcks = []
        refind_stracks = []
        lost_stracks = []
        removed_stracks = []

        remain_inds = scores > self.track_thresh
        inds_low = scores > 0.1
        inds_high = scores < self.track_thresh

        inds_second = np.logical_and(inds_low, inds_high)
        dets_second = bboxes[inds_second]
        dets = bboxes[remain_inds]
        scores_keep = scores[remain_inds]
        scores_second = scores[inds_second]

        detections = []
        if len(dets) > 0:
            detections = [STrack(STrack.tlbr_to_tlwh(tlbr), s) for
                          (tlbr, s) in zip(dets, scores_keep)]

        ''' Add newly detected tracklets to tracked_stracks'''
        unconfirmed = []
        tracked_stracks = []  # type: list[STrack]
        for track in self.tracked_stracks:
            if not track.is_activated:
                unconfirmed.append(track)
            else:
                tracked_stracks.append(track)

        ''' Step 2: First association, with high score detection boxes'''
        strack_pool = joint_stracks(tracked_stracks, self.lost_stracks)
        # Predict the current location with KF
        STrack.multi_predict(strack_pool)
        dists = iou_distance(strack_pool, detections)
        if not self.mot20:
            dists = fuse_score(dists, detections)
        matches, u_track, u_detection = linear_assignment(dists, thresh=self.match_thresh)

        for itracked, idet in matches:
            track = strack_pool[itracked]
            det = detections[idet]
            if track.state == TrackState.Tracked:
                track.update(detections[idet], self.frame_id)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)

        ''' Step 3: Second association, with low score detection boxes'''
        # association the untrack to the low score detections
        if len(dets_second) > 0:
            '''Detections'''
            detections_second = [STrack(STrack.tlbr_to_tlwh(tlbr), s) for
                          (tlbr, s) in zip(dets_second, scores_second)]
        else:
            detections_second = []
        r_tracked_stracks = [strack_pool[i] for i in u_track if strack_pool[i].state == TrackState.Tracked]
        dists = iou_distance(r_tracked_stracks, detections_second)
        matches, u_track, u_detection_second = linear_assignment(dists, thresh=0.5)
        for itracked, idet in matches:
            track = r_tracked_stracks[itracked]
            det = detections_second[idet]
            if track.state == TrackState.Tracked:
                track.update(det, self.frame_id)
                activated_starcks.append(track)
            else:
                track.re_activate(det, self.frame_id, new_id=False)
                refind_stracks.append(track)

        for it in u_track:
            track = r_tracked_stracks[it]
            if not track.state == TrackState.Lost:
                track.mark_lost()
                lost_stracks.append(track)

        '''Deal with unconfirmed tracks, usually tracks with only one beginning frame'''
        detections = [detections[i] for i in u_detection]
        dists = iou_distance(unconfirmed, detections)
        if not self.mot20:
            dists = fuse_score(dists, detections)
        matches, u_unconfirmed, u_detection = linear_assignment(dists, thresh=0.7)
        for itracked, idet in matches:
            unconfirmed[itracked].update(detections[idet], self.frame_id)
            activated_starcks.append(unconfirmed[itracked])
        for it in u_unconfirmed:
            track = unconfirmed[it]
            track.mark_removed()
            removed_stracks.append(track)

        """ Step 4: Init new stracks"""
        for inew in u_detection:
            track = detections[inew]
            if track.score < self.det_thresh:
                continue
            track.activate(self.kalman_filter, self.frame_id)
            activated_starcks.append(track)
        """ Step 5: Update state"""
        for track in self.lost_stracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)

        # print('Ramained match {} s'.format(t4-t3))

        self.tracked_stracks = [t for t in self.tracked_stracks if t.state == TrackState.Tracked]
        self.tracked_stracks = joint_stracks(self.tracked_stracks, activated_starcks)
        self.tracked_stracks = joint_stracks(self.tracked_stracks, refind_stracks)
        self.lost_stracks = sub_stracks(self.lost_stracks, self.tracked_stracks)
        self.lost_stracks.extend(lost_stracks)
        self.lost_stracks = sub_stracks(self.lost_stracks, self.removed_stracks)
        self.removed_stracks.extend(removed_stracks)
        self.tracked_stracks, self.lost_stracks = remove_duplicate_stracks(self.tracked_stracks, self.lost_stracks)
        # get scores of lost tracks
        output_stracks = [track for track in self.tracked_stracks if track.is_activated]

        return output_stracks


def joint_stracks(tlista, tlistb):
    exists = {}
    res = []
    for t in tlista:
        exists[t.track_id] = 1
        res.append(t)
    for t in tlistb:
        tid = t.track_id
        if not exists.get(tid, 0):
            exists[tid] = 1
            res.append(t)
    return res


def sub_stracks(tlista, tlistb):
    stracks = {}
    for t in tlista:
        stracks[t.track_id] = t
    for t in tlistb:
        tid = t.track_id
        if stracks.get(tid, 0):
            del stracks[tid]
    return list(stracks.values())


def remove_duplicate_stracks(stracksa, stracksb):
    pdist = iou_distance(stracksa, stracksb)
    pairs = np.where(pdist < 0.15)
    dupa, dupb = list(), list()
    for p, q in zip(*pairs):
        timep = stracksa[p].frame_id - stracksa[p].start_frame
        timeq = stracksb[q].frame_id - stracksb[q].start_frame
        if timep > timeq:
            dupb.append(q)
        else:
            dupa.append(p)
    resa = [t for i, t in enumerate(stracksa) if not i in dupa]
    resb = [t for i, t in enumerate(stracksb) if not i in dupb]
    return resa, resb


def results_to_arrays(results):
        bboxes, scores = [], []
        for result in results:
            x, y, w, h = result['box']
            bboxes.append([x, y, x + w, y + h])
            scores.append(result['score'])
        return np.array(bboxes), np.array(scores)


class Tracker():
    def __init__(self, track_thresh=0.25, track_buffer=30, match_thresh=0.8, frame_rate=30):
        self.tracker = ByteTracker(track_thresh, track_buffer, match_thresh, frame_rate)

    def update(self, results):
        if results:
            bboxes, scores = results_to_arrays(results)
            tracks = self.tracker.update(bboxes, scores)
            if len(tracks) > 0:
                track_bounding_boxes = np.asarray([track.tlbr for track in tracks])
                ious = box_iou_batch(bboxes, track_bounding_boxes)
                matches, _, _ = linear_assignment(1 - ious, 0.5)
                for i_detection, i_track in matches:
                    results[i_detection]['track_id'] = tracks[i_track].track_id
        return results



import os
import re
import io
import urllib
import requests
import filetype
import itertools
import imagehash

from tqdm import tqdm
from PIL import Image
from transformers import pipeline
from ..utils import HEADERS, load_image, load_url
from . import list_datasets, list_images, load_annotations, save_annotations

GOOGLE_BASE_URL = 'https://www.google.com/search?q='
GOOGLE_PICTURE_ID = '''&biw=1536&bih=674&tbm=isch&sxsrf=ACYBGNSXXpS6YmAKUiLKKBs6xWb4uUY5gA:1581168823770&source=lnms&sa=X&ved=0ahUKEwioj8jwiMLnAhW9AhAIHbXTBMMQ_AUI3QUoAQ'''

BING_BASE_URL = 'https://www.bing.com/images/async?q='


def download_page(url):
    resp = requests.get(url, headers=HEADERS)
    resp.raise_for_status()
    return resp.text


def save_image(link, output_dir, timeout=10, max_size=1920):
    resp = requests.get(link, headers=HEADERS, allow_redirects=True, timeout=timeout)
    kind = filetype.guess(resp.content)
    if kind and kind.mime.startswith('image'):
        d = io.BytesIO(resp.content)
        # d.seek(0)
        im = Image.open(d).convert('RGB')
        im.thumbnail([max_size, max_size])
        phash = str(imagehash.phash(im))
        im.save(os.path.join(output_dir, phash + '.jpg'))
    else:
        raise ValueError(f'Invalid image, not saving')


def get_filter(shorthand):
        if shorthand == "line" or shorthand == "linedrawing":
            return "+filterui:photo-linedrawing"
        elif shorthand == "photo":
            return "+filterui:photo-photo"
        elif shorthand == "clipart":
            return "+filterui:photo-clipart"
        elif shorthand == "gif" or shorthand == "animatedgif":
            return "+filterui:photo-animatedgif"
        elif shorthand == "transparent":
            return "+filterui:photo-transparent"
        else:
            return ""
        

def scan_bing_page(html):
    links = re.findall('murl&quot;:&quot;(.*?)&quot;', html)
    for link in links:
        link = link.replace(" ", "%20")
        yield link

        
def search_bing(query, limit=50, adult='off', filter=''):
    for page_counter in range(100):
        # Parse the page source and download pics
        request_url = BING_BASE_URL + urllib.parse.quote_plus(query) \
                        + '&first=' + str(page_counter) + '&count=' + str(limit) \
                        + '&adlt=' + adult + '&qft=' + get_filter(filter)
        html = download_page(request_url)
        for link in scan_bing_page(html):
            yield link


def scan_google_page(html, extensions={'.jpg', '.jpeg', '.webp'}, timer=5000):
    """Scans for pictures to download based on the keywords"""
    SCANNER_COUNTER = -1
    scanner = html.find
    for _ in range(timer):
        new_line = scanner('"https://', SCANNER_COUNTER + 1)  # How Many New lines
        SCANNER_COUNTER = scanner('"', new_line + 1)  # Ends of line
        buffor = scanner('\\', new_line + 1, SCANNER_COUNTER)
        last_line = buffor if buffor != -1 else SCANNER_COUNTER
        link = html[new_line + 1:last_line]
        if any(extension in link for extension in extensions):
            link = link.replace(" ", "%20")
            yield link


def search_google(query):
    request_url = GOOGLE_BASE_URL + urllib.parse.quote_plus(query) + GOOGLE_PICTURE_ID
    html = download_page(request_url)
    for link in scan_google_page(html):
        yield link


def download(query, limit=100, output_dir='dataset', verbose=True):
    seen = set()
    download_count = 0
    os.makedirs(output_dir, exist_ok=True)
    links = [search_google(query), search_bing(query, adult='off', filter='')]
    ends = [False] * len(links)
    for id in itertools.cycle(range(len(links))):
        try:
            link = next(links[id])
            if link not in seen:
                seen.add(link)
                if download_count < limit:
                    try:
                        save_image(link, output_dir)
                        download_count += 1
                        if verbose:
                            print(f"[%] Downloaded Image #{download_count} from {link}")
                    except Exception as e:
                        print(f"[!] Error getting {link}: {e}")
                else:
                    break
        except StopIteration:
            ends[id] = True
            if set(ends) == {True}:
                break

def search_images(query, limit=100, output_dir='dataset', verbose=True):
    """Search and download images from Google and Bing."""
    download(query, limit=limit, output_dir=output_dir,  verbose=verbose)
    files = [os.path.join(output_dir, f) for f in os.listdir(output_dir) if os.path.isfile(os.path.join(output_dir, f))]
    return files


# As the Grounding DINO model was trained with a "." after each text, we'll do the same here.
def preprocess_caption(caption: str) -> str:
    result = caption.lower().strip()
    return result if result.endswith('.') else result + '.'


def format_results(results, threshold=0.6):
    r = []
    for result in results:
        score = result["score"]
        if score > threshold:
            label = result["label"].rpartition('.')[0]
            xmin, ymin, xmax, ymax = result['box'].values()
            r.append({"label": label, "score": score, "box": [xmin, ymin, xmax - xmin, ymax - ymin]})
    return r


def annotate_image(pipe, img, classes, threshold=0.3):
    im = Image.fromarray(img)
    labels = [preprocess_caption(txt) for txt in classes]
    objects = format_results(pipe(im, candidate_labels=labels, threshold=threshold))
    return objects


def annotate_images(images, classes):
    """Annotate a dataset using Grounding Dino."""
    pipe = pipeline(task="zero-shot-object-detection", model="IDEA-Research/grounding-dino-tiny")
    #pipe = pipeline(task="zero-shot-object-detection", model="google/owlv2-base-patch16-ensemble")
    annotations = []
    for row in tqdm(images):
        url, filename = row['url'], row['name']
        img = load_image(load_url(url))
        objects = annotate_image(pipe, img, classes)
        if objects:
            annotation = {'url': url, 'filename': filename, 'objects': objects}
            annotations.append(annotation)
    return annotations



import pathlib
import pkg_resources

from setuptools import setup, find_packages

from abraia import __version__

with open('README.md') as f:
    long_description = f.read()

with pathlib.Path('requirements.txt').open() as requirements_txt:
    install_requires = [str(requirement) for requirement
        in pkg_resources.parse_requirements(requirements_txt)]
    
extras_require = {
    'multiple': ['spectral>=0.23.1', 'scipy>=1.14.1', 'tifffile>=2020.9.4'],
    'dev': ['opencv-python>=4.10.0.84', 'ultralytics>=8.3.59', 'onnx>=1.18.0'],#, 'onnxsim>=0.4.36'],
}

setup(
    name='abraia',
    version=__version__,
    description='Abraia Python SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/abraia/abraia-multiple',
    author='Jorge Rodriguez Araujo',
    author_email='jorge@abraiasoftware.com',
    license='MIT',
    zip_safe=False,
    packages=find_packages(exclude=['tests']),
    package_data={'': ['*.jpg', '*.gz']},
    include_package_data=True,
    tests_require=['pytest'],
    setup_requires=['setuptools>=38.6.0', 'pytest-runner'],
    scripts=['scripts/abraia', 'scripts/abraia.bat'],
    install_requires=install_requires,
    extras_require=extras_require
)



import numpy as np
import onnxruntime as ort

from ..utils import download_file

sess_options = ort.SessionOptions()
sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL


class SwinIR:
    def __init__(self):
        model_src = download_file('multiple/models/editing/003_realSR_BSRGAN_DFO_s64w8_SwinIR-M_x4_GAN.onnx')
        self.session = ort.InferenceSession(model_src, sess_options)
        self.input_name = self.session.get_inputs()[0].name

    def preprocess(self, img):
        img = (img / 255).astype(np.float32)
        return np.expand_dims(img.transpose((2, 0, 1)), axis=0)

    def postprocess(self, img):
        img = (255 * img.clip(0, 1)).astype(np.uint8)
        return img.transpose((1, 2, 0))

    def upscale(self, img):
        inputs = {self.input_name: self.preprocess(img)}
        outputs = self.session.run(None, inputs)
        return self.postprocess(outputs[0][0])


def create_gradient_mask(shape, feather):
    """Create a gradient mask for smooth blending of tiles."""
    mask = np.ones(shape)
    _, _, h, w = shape
    for feather_step in range(feather):
        factor = (feather_step + 1) / feather
        mask[:, :, feather_step, :] *= factor
        mask[:, :, h - 1 - feather_step, :] *= factor
        mask[:, :, :, feather_step] *= factor
        mask[:, :, :, w - 1 - feather_step] *= factor
    return mask


def tiled_upscale(samples, function, scale, tile_size, overlap = 8):
    """Apply a scaling function to image samples in a tiled manner."""
    height, width = samples.shape[2:]
    tile_width, tile_height = tile_size
    out_height, out_width = round(height * scale), round(width * scale)
    output = np.empty((1, 3, out_height, out_width))
    out = np.zeros((1, 3, out_height, out_width))
    out_div = np.zeros_like(output)
    for y in range(0, height, tile_height - overlap):
        for x in range(0, width, tile_width - overlap):
            # Ensure we don't go out of bounds
            x_end = min(x + tile_width, width)
            y_end = min(y + tile_height, height)
            x = max(0, x_end - tile_width)
            y = max(0, y_end - tile_height)
            # Extract and process the tile
            tile = samples[:, :, y:y_end, x:x_end]
            processed_tile = function(tile)
            # Calculate the position in the output tensor
            out_y, out_x = round(y * scale), round(x * scale)
            out_h, out_w = processed_tile.shape[2:]
            # Create a feathered mask for smooth blending
            mask = create_gradient_mask(processed_tile.shape, overlap * scale)
            # Add the processed tile to the output
            out[:, :, out_y : out_y + out_h, out_x : out_x + out_w] += processed_tile * mask
            out_div[:, :, out_y : out_y + out_h, out_x : out_x + out_w] += mask
    # Normalize the output
    output = out / out_div
    return output


class Upscaler:
    def __init__(self, overlap = 8):
        self.scale = 2
        self.overlap = overlap
        self.tile_size = (1024, 1024)
        # model_src = download_file('multiple/models/editing/4xNomosWebPhoto_RealPLKSR_fp32_opset17.onnx')
        # model_src = download_file('multiple/models/editing/4xNomosUni_span_multijpg_fp32_opset17.onnx')
        model_src = download_file('multiple/models/editing/2xNomosUni_compact_multijpg_ldl_fp32_opset17.onnx')
        self.session = ort.InferenceSession(model_src, sess_options)
        self.input_name = self.session.get_inputs()[0].name

    def preprocess(self, img):
        img = (img / 255).astype(np.float32)
        return np.expand_dims(img.transpose((2, 0, 1)), axis=0)

    def postprocess(self, img):
        img = (255 * img.clip(0, 1)).astype(np.uint8)
        return img.transpose((1, 2, 0))
    
    def predict(self, img):
        return self.session.run(None, {self.input_name: img})[0]

    def upscale(self, img):
        img = self.preprocess(img)
        outputs = tiled_upscale(img, self.predict, self.scale, self.tile_size, self.overlap)
        return self.postprocess(outputs[0])
